library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- To use the utility library (clk generator, checkers, alerts)
library uvvm_util;
use uvvm_util.types_pkg.all;
use uvvm_util.string_methods_pkg.all;
use uvvm_util.adaptations_pkg.all;
use uvvm_util.methods_pkg.all;

-- To use randomization and functional coverage
library osvvm;
use osvvm.RandomBasePkg.all;
use osvvm.RandomPkg.all;
use osvvm.CoveragePkg.all;


entity tb_bram is
  generic (
    DATA_WIDTH : integer := 8;
    ADDR_WIDTH : integer := 8
    );
end tb_bram;

architecture simulation of tb_bram is

  -- DUT I/Os
  signal clk   : std_logic;
  signal addri : unsigned (ADDR_WIDTH-1 downto 0);
  signal datai : std_logic_vector (DATA_WIDTH-1 downto 0);
  signal we    : std_logic;
  signal datao : std_logic_vector (DATA_WIDTH-1 downto 0);

  -- Expected DUT state
  type ram_type is array ((2**ADDR_WIDTH)-1 downto 0) of std_logic_vector (DATA_WIDTH-1 downto 0);
  signal expected : ram_type;

  -- Clock Control
  constant CLK_PERIOD : time    := 10 ns;
  signal   clk_enable : boolean := true;

begin

  -- This way of instancing doesn't require declaring the component
  bram_inst: entity work.bram
    generic map (
      DATA_WIDTH => DATA_WIDTH,
      ADDR_WIDTH => ADDR_WIDTH
      )
    port map (
      clk => clk,
      addri => addri,
      datai => datai,
      we => we,
      datao => datao
      );


  -- Clock generator
  clock_generator (clk, clk_enable, CLK_PERIOD, "Global clock",
                   CLK_PERIOD/2);

  -- Main testbench process
  uvvm_main: process

    -- Random number generation
    variable RV : RandomPTYpe;

    -- Drive the memory: write data to addr
    procedure write (constant data: in integer;
                     constant addr: in integer) is
    begin
      datai <= std_logic_vector(to_unsigned(data, DATA_WIDTH));
      addri <= to_unsigned(addr, ADDR_WIDTH);
      we    <= '1';
      wait for clk_period;
      we    <= '0';
    end procedure;

    -- Drive the memory: read data from addr
    procedure read (constant addr: in integer) is
    begin
      addri <= to_unsigned(addr, ADDR_WIDTH);
      wait for clk_period;
    end procedure;

  begin

    -- Tell the testbench to stop if a number of errors is found
    set_alert_stop_limit(ERROR, 10);    -- Stop on a number of errors
    set_alert_stop_limit(TB_ERROR, 2);  -- Stop on a number of testbench errors

    -- Disable positive acknowledgements ("OK" messages)
    --disable_log_msg(ID_POS_ACK);

    -- Start simulation
    log (ID_LOG_HDR, "Starting simulation");
    clk_enable <= true;

    -- Wait for a bit of time before starting to put data in the memory: if we
    -- don't, the predictor process may miss the first data because of a race
    -- condition (suppose the predictor process is executed at time 0, before
    -- the uvvm process: it will not catch the read transaction)
    wait for 0.5*clk_period;

    -- Fill the BRAM with random data
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      write ( RV.RandInt(0, (2**DATA_WIDTH)-1 ), i);
    end loop;

    -- Read all data and check read data with the expected values
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      read(i);
      check_value(datao, expected(i), ERROR, "Check if read data matches expected");
    end loop;

    -- After all tests, end simulation
    log (ID_LOG_HDR, "Simulation completed");
    clk_enable <= false;

    -- Report summary of alerts
    report_alert_counters(VOID);

    wait;

  end process;

  -- Predictor: catch write operations and predict the internal state of the memory
  predictor: process (clk)
  begin
    if (rising_edge(clk) and we = '1') then
      expected(to_integer(addri)) <= datai;
    end if;
  end process;

end simulation;
