library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

-- To use the utility library (clk generator, checkers, alerts)
library uvvm_util;
use uvvm_util.types_pkg.all;
use uvvm_util.string_methods_pkg.all;
use uvvm_util.adaptations_pkg.all;
use uvvm_util.methods_pkg.all;

-- To use randomization and functional coverage
library osvvm;
use osvvm.RandomBasePkg.all;
use osvvm.RandomPkg.all;
use osvvm.CoveragePkg.all;


entity tb_dpram is
  generic (
    DATA_WIDTH : integer := 8;
    ADDR_WIDTH : integer := 8
    );
end tb_dpram;

architecture simulation of tb_dpram is

  -- DUT I/Os
  signal clk_a   : std_logic;
  signal addri_a : unsigned (ADDR_WIDTH-1 downto 0);
  signal datai_a : std_logic_vector (DATA_WIDTH-1 downto 0);
  signal we_a    : std_logic;
  signal datao_a : std_logic_vector (DATA_WIDTH-1 downto 0);
  signal clk_b   : std_logic;
  signal addri_b : unsigned (ADDR_WIDTH-1 downto 0);
  signal datai_b : std_logic_vector (DATA_WIDTH-1 downto 0);
  signal we_b    : std_logic;
  signal datao_b : std_logic_vector (DATA_WIDTH-1 downto 0);

  -- Expected DUT state
  type ram_type is array ((2**ADDR_WIDTH)-1 downto 0) of std_logic_vector (DATA_WIDTH-1 downto 0);
  signal expected : ram_type;

  -- Clock Control
  constant CLK_PERIOD : time      := 10 ns;
  signal   clk_enable : boolean   := true;
  signal   clk        : std_logic := '0';

begin

  -- This way of instancing doesn't require declaring the component
  dpram_inst: entity work.dpram
    generic map (
      DATA_WIDTH => DATA_WIDTH,
      ADDR_WIDTH => ADDR_WIDTH
      )
    port map (
      clk_a   => clk_a,
      addri_a => addri_a,
      datai_a => datai_a,
      we_a    => we_a,
      datao_a => datao_a,
      clk_b   => clk_b,
      addri_b => addri_b,
      datai_b => datai_b,
      we_b    => we_b,
      datao_b => datao_b
      );


  -- Clock generator
  clock_generator (clk, clk_enable, CLK_PERIOD, "Global clock", CLK_PERIOD/2);
  clk_a <= clk;
  clk_b <= clk;

  -- Main testbench process
  uvvm_main: process

    -- Random number generation
    variable RV : RandomPTYpe;

    -- Drive the memory: write data to addr
    procedure write_a (constant data: in integer;
                       constant addr: in integer) is
    begin
      datai_a <= std_logic_vector(to_unsigned(data, DATA_WIDTH));
      addri_a <= to_unsigned(addr, ADDR_WIDTH);
      we_a    <= '1';
      wait for clk_period;
      we_a    <= '0';
    end procedure;

    procedure write_b (constant data: in integer;
                       constant addr: in integer) is
    begin
      datai_b <= std_logic_vector(to_unsigned(data, DATA_WIDTH));
      addri_b <= to_unsigned(addr, ADDR_WIDTH);
      we_b    <= '1';
      wait for clk_period;
      we_b    <= '0';
    end procedure;

    -- Drive the memory: read data from addr
    procedure read_a (constant addr: in integer) is
    begin
      addri_a <= to_unsigned(addr, ADDR_WIDTH);
      wait for clk_period;
    end procedure;

    procedure read_b (constant addr: in integer) is
    begin
      addri_b <= to_unsigned(addr, ADDR_WIDTH);
      wait for clk_period;
    end procedure;

  begin

    -- Tell the testbench to stop if a number of errors is found
    set_alert_stop_limit(ERROR, 10);    -- Stop on a number of errors
    set_alert_stop_limit(TB_ERROR, 2);  -- Stop on a number of testbench errors

    -- Uncomment to disable positive acknowledgements ("OK" messages)
    --disable_log_msg(ID_POS_ACK);

    -- Start simulation
    log (ID_LOG_HDR, "Starting simulation");
    clk_enable <= true;

    -- Wait for a bit of time before starting to put data in the memory: if we
    -- don't, the predictor process may miss the first data because of a race
    -- condition (suppose the predictor process is executed at time 0, before
    -- the uvvm process: it will not catch the read transaction)
    wait for 0.5*clk_period;

    -- Fill the BRAM with random data, using port a
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      write_a ( RV.RandInt(0, (2**DATA_WIDTH)-1 ), i);
    end loop;

    -- Read all data and check read data with the expected values, using port a
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      read_a(i);
      check_value(datao_a, expected(i), ERROR, "Check if read data matches expected (port a)");
    end loop;

    -- Read all data and check read data with the expected values, using port b
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      read_b(i);
      check_value(datao_b, expected(i), ERROR, "Check if read data matches expected (port b)");
    end loop;


    -- Fill the BRAM with random data again, this time using port b
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      write_b ( RV.RandInt(0, (2**DATA_WIDTH)-1 ), i);
    end loop;

    -- Read all data and check read data with the expected values, using port a
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      read_a(i);
      check_value(datao_a, expected(i), ERROR, "Check if read data matches expected (port a)");
    end loop;

    -- Read all data and check read data with the expected values, using port b
    for i in 0 to (2**ADDR_WIDTH)-1 loop
      read_b(i);
      check_value(datao_b, expected(i), ERROR, "Check if read data matches expected (port b)");
    end loop;



    -- After all tests, end simulation
    log (ID_LOG_HDR, "Simulation completed");
    clk_enable <= false;

    -- Report summary of alerts
    report_alert_counters(VOID);

    wait;

  end process;

  -- Predictor: catch write operations and predict the internal state of the memory
  predictor: process (clk)
  begin
      if rising_edge(clk) then
        if (we_a = '1') then
            expected(to_integer(addri_a)) <= datai_a;
        end if;
        if (we_b = '1') then
            expected(to_integer(addri_b)) <= datai_b;
        end if;
    end if;
  end process;

end simulation;
