LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

-- To use the utility library (clk generator, checkers, alerts)
library uvvm_util;
use uvvm_util.types_pkg.all;
use uvvm_util.string_methods_pkg.all;
use uvvm_util.adaptations_pkg.all;
use uvvm_util.methods_pkg.all;

-- To use randomization and functional coverage
--
-- Interestingly, OSVVM uses CamelCase while UVVM uses underscore_case
-- Since VHDL is case insensitive, we could use all OSVVM functionality in
-- lowercase if we wanted
library osvvm;
use osvvm.RandomBasePkg.all;
use osvvm.RandomPkg.all;
use osvvm.CoveragePkg.all;

ENTITY tb_divider IS
  Generic(
    N : integer := 8;               -- Dividend width
    M : integer := 8;               -- Divisor width
    DIRECTED   : boolean := true;  -- Do a basic directed test
    NUM_RANDOM : integer := 2000    -- Number of random inputs
    );
END tb_divider;

ARCHITECTURE behavior OF tb_divider IS

  -- Component Declaration for the Unit Under Test (UUT)

  COMPONENT divider
    Generic (N : integer := 8;  -- Number of bits in dividend and quotient
             M : integer := 8); -- Number of bits in divisor and remainder
    PORT(
      dividend : IN  signed(N-1 downto 0);
      divisor : IN  signed(M-1 downto 0);
      valid : IN  std_logic;
      quotient : OUT  signed(N-1 downto 0);
      remainder : OUT  signed(M-1 downto 0);
      out_valid : OUT  std_logic;
      busy : OUT  std_logic;
      err : OUT  std_logic;
      rst : IN  std_logic;
      clk : IN  std_logic
      );
  END COMPONENT;

  --Inputs
  signal dividend : signed(N-1 downto 0) := (others => '0');
  signal divisor : signed(M-1 downto 0) := (others => '0');
  signal valid : std_logic := '0';
  signal rst : std_logic := '0';
  signal clk : std_logic := '0';

  --Outputs
  signal quotient : signed(N-1 downto 0);
  signal remainder : signed(M-1 downto 0);
  signal out_valid : std_logic;
  signal busy : std_logic;
  signal err : std_logic;

  -- Clock control
  signal clk_enable : boolean := true;
  constant clk_period : time := 10 ns;

    -- Declare Coverage Objects (for functional coverage)
    shared variable DividendCover : CovPType;
    shared variable DivisorCover  : CovPType;
    shared variable CrossCover    : CovPtype;

BEGIN

  assert (N >= 0) and (N <= 32)
    report "tb_divider: N (dividend and quotient width) must be between 0 and 32 bits. For N > 32 bits the hardware may be correct, but since the testbench uses 32-bit integers we cannot check it without changing the testbench"
    severity failure;

  assert (M >= 0) and (M <= 32)
    report "tb_divider: M (divisor and remainder width) must be between 0 and 32 bits. For M > 32 bits the hardware may be correct, but since the testbench uses 32-bit integers we cannot check it without changing the testbench"
    severity failure;

  -- Instantiate the Unit Under Test (UUT)
  uut: divider
    Generic MAP (
      N => N,
      M => M
      )
    PORT MAP (
      dividend => dividend,
      divisor => divisor,
      valid => valid,
      quotient => quotient,
      remainder => remainder,
      out_valid => out_valid,
      busy => busy,
      err => err,
      rst => rst,
      clk => clk
      );

  -- Clock generator
  clock_generator (clk, clk_enable, CLK_PERIOD, "Global clock", CLK_PERIOD/2);

  -- Main testbench process
  uvvm_main: process

    -- Random number generation
    variable RV : RandomPTYpe;

    -- Procedure to stimulate the divider
    -- Procedures that directly modify signals or variables must be defined
    -- inside a process. By defining the procedure inside the process we
    -- avoid having procedure parameters for inf, sup and valid, which
    -- makes it easier to call
    procedure drive (constant A, B: in integer) is
    begin
      -- Assert that the inputs are in the correct range
      -- -(2^(N-1)),2^(N-1)-1 for dividend
      -- -(2^(M-1)),2^(M-1)-1 for divisor
      check_value_in_range(A, -(2**(N-1)), 2**(N-1)-1, TB_FAILURE, "dividend in range", "driver");
      check_value_in_range(B, -(2**(M-1)), 2**(M-1)-1, TB_FAILURE, "divisor in range", "driver");

      -- Drive the divider pins
      valid    <= '1';
      dividend <= to_signed(A,N);
      divisor  <= to_signed(B,M);
      wait for 1*clk_period;
      valid <= '0';
      -- We know the state machine takes N+1 clock cycles to finish, but we
      -- could also wait for out_valid
      wait for clk_period*(N+1);
    end procedure;

  begin
    set_alert_stop_limit(ERROR, 2);    -- Stop on a number of errors
    set_alert_stop_limit(TB_ERROR, 1);  -- Stop on a number of testbench errors

    -- Uncomment this to disable "OK" messages (POSitive ACKnowledgements)
    --disable_log_msg(ID_POS_ACK);

    -- Enable clock
    log (ID_LOG_HDR, "Start simulation");
    clk_enable <= true;

    -- Test sequence defined here

    -- Reset the DUT and set all inputs to inactive values
    rst <= '1';
    valid <= '0';
    dividend <= (others => '0');
    divisor <= (others => '0');
    wait for 1*clk_period;
    rst <= '0';

    -- Directed test
    if (DIRECTED = true) then
      drive (100, 9);
      drive (120, -7);
      drive (0, 50);
      drive (0, -50);
    end if;

    -- Generate some random data:
    -- 1.- Initialize random seed
    -- 2.- Drive N times with random data. Exclude 0 from the possible values for divisor!
    RV.InitSeed(RV'instance_name);

    for i in 1 to NUM_RANDOM loop
      drive ( RV.RandInt( -(2**(N-1)), (2**(N-1))-1) , RV.RandInt( -(2**(M-1)), (2**(M-1))-1, (0,0)) );
    end loop;

    -- After all tests, end simulation
    log (ID_LOG_HDR, "Simulation completed");
    clk_enable <= false;

        -- Report functional coverage.
        log (ID_LOG_HDR, "Reporting functional coverage");
        DividendCover.WriteBin;
        DivisorCover.WriteBin;
        CrossCover.WriteBin;

        log (ID_LOG_HDR, "Reporting coverage holes");
        DividendCover.WriteCovHoles;
        DivisorCover.WriteCovHoles;
        CrossCover.WriteCovHoles;

        -- Check for any coverage holes. Notice the use of TB_ERROR instead of
        -- ERROR, to indicate the problem is in the testbench, not in the DUT
        log (ID_LOG_HDR, "Checking for coverage holes");
        check_value(DividendCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in Dividend coverage");
        check_value(DivisorCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in Divisor coverage");
        check_value(CrossCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in cross coverage");

    -- Report summary of alerts
    report_alert_counters(VOID);

    -- Avoid repeating the test indefinitely (do not reach "end process")
    wait;

  end process;

  -- Monitor that the outputs are correct
  -- Here we will use the fact that the driver will not change the inputs
  -- between transactions, so when we get an out_data = valid, output quotient
  -- and reminder should match with the current dividend and divisor
  protocol_monitor: process (clk)

      variable expected_quotient : signed(N-1 downto 0);
      variable expected_remainder : signed(M-1 downto 0);

  begin
    if falling_edge(clk) then
      if (out_valid /= '0') then
        -- Calculate the expected output data
        expected_quotient := dividend/divisor;
        expected_remainder := to_signed(integer(real(to_integer(dividend)) -
        (real(to_integer(divisor)) * real(to_integer(quotient)))), M);
        -- Compare actual to expected
        check_value (quotient, expected_quotient, ERROR, "Check quotient", "monitor");
        check_value (remainder, expected_remainder, ERROR, "Check remainder", "monitor");
      end if;
    end if;
  end process;

    -- Collect Coverage
    -- AddBins(): add item coverage bin(s) to internal data structure
    -- GenBin(min, max, [numbins]): create array of bin inputs for AddBins
    --
    -- What to do in this process is not obvious, but is described in OSVVM
    -- documentation:
    --   1.- Optionally set a name for each coverage object
    --   2.- Add bins to each coverage object
    --   3.- Record data in the coverage objects each time we have a valid data
    CollectCov: process
    begin

        -- Set names for Coverage Objects
        --DividendCover.SetName( "Cover inf.re" );
        DividendCover.SetName("Functional Coverage for Dividend --------------");
        DivisorCover.SetName ("Functional Coverage for Divisor  --------------");
        CrossCover.SetName   ("Cross Coverage for Dividend, Divisor ----------");

        -- Define functional coverage bins
        -- We can divide the positive/negative ranges in multiple bins just
        -- by changing the third argument of GenBins
        DividendCover.AddBins( "inf.re", GenBin (          1, (2**(N-1))-1, 1) );  -- Positive values
        DividendCover.AddBins( "inf.re", GenBin (          0,            0, 1) );  -- Special bin for zero
        DividendCover.AddBins( "inf.re", GenBin (-(2**(N-1)),           -1, 1) );  -- Negative values

        DivisorCover.AddBins( "inf.re", GenBin (          1, (2**(M-1))-1, 1) );  -- Positive values
        --DivisorCover.AddBins( "inf.re", GenBin (          0,            0, 1));  -- Special bin for zero (should not happen!)
        DivisorCover.AddBins( "inf.re", GenBin (-(2**(M-1)),           -1, 1) );  -- Negative values

        -- Define cross coverage bins
        -- For each of the two inputs, define a number of bins. All the
        -- possibilities will be combined to generate a huge number of bins
        CrossCover.AddCross (
            GenBin ( 1,  (2**(N-1))-1, 1) & GenBin (0, 0, 1) &  GenBin (-(2**(N-1)), -1, 1),  -- Dividend's three bins
            GenBin ( 1,  (2**(M-1))-1, 1) &  GenBin (-(2**(M-1)), -1, 1)   -- Divisor's two bins
        );

        -- Record the data in the Coverage Objects every time we have a valid input
        while clk_enable = true loop

            wait on clk until clk = '1' and valid = '1';
            DividendCover.ICover(to_integer(dividend));
            DivisorCover.ICover(to_integer(divisor));
            CrossCover.ICover((to_integer(dividend), to_integer(divisor)));

        end loop;

    end process;

END;
