library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.edc_common.all;

-- To use the utility library (clk generator, checkers, alerts)
library uvvm_util;
use uvvm_util.types_pkg.all;
use uvvm_util.string_methods_pkg.all;
use uvvm_util.adaptations_pkg.all;
use uvvm_util.methods_pkg.all;

-- To use randomization and functional coverage
--
-- Interestingly, OSVVM uses CamelCase while UVVM uses underscore_case
-- Since VHDL is case insensitive, we could use all OSVVM functionality in
-- lowercase if we wanted
library osvvm;
use osvvm.RandomBasePkg.all;
use osvvm.RandomPkg.all;
use osvvm.CoveragePkg.all;

entity tb_interpolator is
    GENERIC (
        DIRECTED   : boolean := true;   -- Do a basic directed test
        NUM_RANDOM : integer := 2000;   -- Number of random inputs
        EXHAUSTIVE : boolean := false   -- Do an exhaustive simulation between MIN and MAX values (defined later)
    );
end tb_interpolator;

architecture software of tb_interpolator is

    -- I/Os
    signal rst, clk           : std_logic;
    signal valid, estim_valid : std_logic;
    signal inf, sup, estim    : complex10 := (re => to_signed(0,10), im => to_signed(0, 10));

    -- clk control
    signal clk_enable   : boolean := true;
    constant CLK_PERIOD : time := 10 ns;

    -- protocol monitor
    signal monitor_i      : integer := 12;
    signal expected_re    : real;
    signal expected_im    : real;
    signal expected       : complex10;

    -- MIN/MAX values for exhaustive verification
    -- If MIN and MAX cover all the 10-bit range, simulation will take a HUGE
    -- time to finish (2^40: best to use a workstation over a weekend and
    -- disable VCD/GHW generation)
    constant MIN : integer := -4; -- -(2**9);
    constant MAX : integer :=  4; -- (2**9)-1;

    -- Declare Coverage Objects (for functional coverage)
    shared variable InfReCover : CovPType;
    shared variable InfImCover : CovPType;
    shared variable SupReCover : CovPType;
    shared variable SupImCover : CovPType;
    shared variable CrossCover : CovPType;

begin

    -- Unit under test
    interpolator_inst: entity work.interpolator
    port map (
                rst         => rst,
                clk         => clk,
                inf         => inf,
                sup         => sup,
                valid       => valid,
                estim       => estim,
                estim_valid => estim_valid
             );

    -- Clock generator
    clock_generator (clk, clk_enable, CLK_PERIOD, "Global clock", CLK_PERIOD/2);

    -- Main testbench process
    uvvm_main: process

        -- Random number generation
        variable RV : RandomPTYpe;

        -- Procedure to stimulate the interpolator
        -- Procedures that directly modify signals or variables must be defined
        -- inside a process. By defining the procedure inside the process we
        -- avoid having procedure parameters for inf, sup and valid, which
        -- makes it easier to call
        procedure drive (constant inf_re, inf_im, sup_re, sup_im: in integer) is
        begin
            -- Assert that the inputs are in the range -512, 511
            check_value_in_range(inf_re, -512, 511, TB_FAILURE, "inf_re in range", "driver");
            check_value_in_range(inf_im, -512, 511, TB_FAILURE, "inf_im in range", "driver");
            check_value_in_range(sup_re, -512, 511, TB_FAILURE, "sup_re in range", "driver");
            check_value_in_range(sup_im, -512, 511, TB_FAILURE, "sup_im in range", "driver");

            -- Drive the interpolator pins
            valid <= '1';
            inf.re <= to_signed(inf_re, 10);
            inf.im <= to_signed(inf_im, 10);
            sup.re <= to_signed(sup_re, 10);
            sup.im <= to_signed(sup_im, 10);
            wait for 1*clk_period;
            valid <= '0';
            wait for 12*clk_period;
        end procedure;

    begin
        set_alert_stop_limit(ERROR, 100);    -- Stop on a number of errors
        set_alert_stop_limit(TB_ERROR, 10);  -- Stop on a number of testbench errors

        -- Uncomment this to disable "OK" messages (POSitive ACKnowledgements)
        --disable_log_msg(ID_POS_ACK);

        -- Enable clock
        log (ID_LOG_HDR, "Start simulation");
        clk_enable <= true;

        -- Test sequence defined here
        rst <= '1';
        valid <= '0';
        inf <= (re => (others=>'0'),     im => (others=>'0'));
        sup <= (re => (others=>'0'),     im => (others=>'0'));
        wait for 1*clk_period;
        rst <= '0';

        -- Directed test
        if (DIRECTED = true) then
            drive (0, 0, 500, -500);
            drive (0, 0, 250, -250);
        end if;

        -- Generate some random data:
        -- 1.- Initialize random seed
        -- 2.- Drive N times with random data
        RV.InitSeed(RV'instance_name);

        for i in 1 to NUM_RANDOM loop
            drive (RV.RandInt(-512,511), RV.RandInt(-512,511), RV.RandInt(-512,511), RV.RandInt(-512,511));
        end loop;

        -- Are we paranoid enough?
        -- Exhaustively verify all possibilities, making inf.re, inf.im,
        -- sup.re, sup.inf take all possible value combinations from MIN to MAX.
        if (EXHAUSTIVE = true) then
            for i in MIN to MAX loop
                for j in MIN to MAX loop
                    for k in MIN to MAX loop
                        for l in MIN to MAX loop
                            drive(i, j, k, l);
                        end loop;
                    end loop;
                end loop;
            end loop;
        end if;

        -- After all tests, end simulation
        log (ID_LOG_HDR, "Simulation completed");
        clk_enable <= false;

        -- Report functional coverage.
        log (ID_LOG_HDR, "Reporting functional coverage");
        InfReCover.WriteBin;
        InfImCover.WriteBin;
        SupReCover.WriteBin;
        SupReCover.WriteBin;

        log (ID_LOG_HDR, "Reporting coverage holes");
        InfReCover.WriteCovHoles;
        InfImCover.WriteCovHoles;
        SupReCover.WriteCovHoles;
        SupReCover.WriteCovHoles;

        -- WriteBin/WriteCovHoles can receive a file name as an argument, in
        -- that case there will be no output on the console. Let's use this
        -- option for cross coverage, since it has by definition lots of bins
        log (ID_LOG_HDR, "Writing cross coverage to coverage.log");
        CrossCover.WriteBin("coverage.log");
        CrossCover.WriteCovHoles("coverage.log");

        -- Check for any coverage holes. Notice the use of TB_ERROR instead of
        -- ERROR, to indicate the problem is in the testbench, not in the DUT
        log (ID_LOG_HDR, "Checking for coverage holes");
        check_value(InfReCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in inf.re coverage");
        check_value(InfImCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in inf.im coverage");
        check_value(SupReCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in sup.re coverage");
        check_value(SupReCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in sup.re coverage");
        check_value(CrossCover.CountCovHoles(100.0), 0, TB_ERROR, "Check for holes in cross coverage");

        -- Report summary of alerts
        report_alert_counters(VOID);

        -- Avoid repeating the test indefinitely (do not reach "end process")
        wait;

    end process;

    -- Monitor that the outputs are correct
    protocol_monitor: process (clk)
    begin
        if falling_edge(clk) then
            if (valid /= '0') then
                if (monitor_i = 12) then
                    monitor_i <= 0;
                else
                    alert (ERROR, "Valid received before expected");
                end if;
            end if;
            if (estim_valid /= '0') then
                -- We have to use a range because the expected values may be
                -- off 1 LSB, because of truncating
                check_value_in_range (estim.re, expected.re-1, expected.re, ERROR, "Check estimated data (real)", "monitor");
                check_value_in_range (estim.im, expected.im-1, expected.im, ERROR, "Check estimated data (imag)", "monitor");
                monitor_i <= monitor_i + 1;
            end if;
        end if;
    end process;

    -- Calculate the expected interpolated data in a different way than in the
    -- interpolator. We don't want to replicate that same code!
    -- So: compute the equation of the straight line and scale by the scale
    -- factor (*12 because we estimate 12*y, and /16 because we drop the 4 LSB)
    expected_re <= (real(12)/real(16))*((real(to_integer(sup.re) - to_integer(inf.re))/real(12))*real(monitor_i) + real(to_integer(inf.re)));
    expected_im <= (real(12)/real(16))*((real(to_integer(sup.im) - to_integer(inf.im))/real(12))*real(monitor_i) + real(to_integer(inf.im)));
    expected.re <= to_signed(integer((expected_re)), 10);
    expected.im <= to_signed(integer((expected_im)), 10);

    -- Collect Coverage
    -- AddBins(): add item coverage bin(s) to internal data structure
    -- GenBin(min, max, [numbins]): create array of bin inputs for AddBins
    --
    -- What to do in this process is not obvious, but is described in OSVVM
    -- documentation:
    --   1.- Optionally set a name for each coverage object
    --   2.- Add bins to each coverage object
    --   3.- Record data in the coverage objects each time we have a valid data
    CollectCov: process
    begin

        -- Set names for Coverage Objects
        --InfReCover.SetName( "Cover inf.re" );
        InfReCover.SetName("Functional Coverage for inf.re --------------");
        InfImCover.SetName("Functional Coverage for inf.im --------------");
        SupReCover.SetName("Functional Coverage for sup.re --------------");
        SupImCover.SetName("Functional Coverage for sup.im --------------");
        CrossCover.SetName("Cross Coverage for inf, sup -----------------");

        -- Define functional coverage bins
        -- We can divide the positive/negative ranges in multiple bins just
        -- by changing the third argument of GenBins
        InfReCover.AddBins( "inf.re", GenBin ( 511,  511, 1) );  -- Special bin for 511
        InfReCover.AddBins( "inf.re", GenBin (   1,  510, 2) );  -- Positive values (other than 511)
        InfReCover.AddBins( "inf.re", GenBin (   0,    0, 1) );  -- Special bin for zero
        InfReCover.AddBins( "inf.re", GenBin (-511,   -1, 2) );  -- Negative values (other than -512)
        InfReCover.AddBins( "inf.re", GenBin (-512, -512, 1) );  -- Special value for 511

        InfImCover.AddBins( "inf.im", GenBin ( 511,  511, 1) );  -- Special bin for 511
        InfImCover.AddBins( "inf.im", GenBin (   1,  510, 2) );  -- Positive values (other than 511)
        InfImCover.AddBins( "inf.im", GenBin (   0,    0, 1) );  -- Special bin for zero
        InfImCover.AddBins( "inf.im", GenBin (-511,   -1, 2) );  -- Negative values (other than -512)
        InfImCover.AddBins( "inf.im", GenBin (-512, -512, 1) );  -- Special value for 511

        SupReCover.AddBins( "sup.re", GenBin ( 511,  511, 1) );  -- Special bin for 511
        SupReCover.AddBins( "sup.re", GenBin (   1,  510, 2) );  -- Positive values (other than 511)
        SupReCover.AddBins( "sup.re", GenBin (   0,    0, 1) );  -- Special bin for zero
        SupReCover.AddBins( "sup.re", GenBin (-511,   -1, 2) );  -- Negative values (other than -512)
        SupReCover.AddBins( "sup.re", GenBin (-512, -512, 1) );  -- Special value for 511

        SupImCover.AddBins( "sup.im", GenBin ( 511,  511, 1) );  -- Special bin for 511
        SupImCover.AddBins( "sup.im", GenBin (   1,  510, 2) );  -- Positive values (other than 511)
        SupImCover.AddBins( "sup.im", GenBin (   0,    0, 1) );  -- Special bin for zero
        SupImCover.AddBins( "sup.im", GenBin (-511,   -1, 2) );  -- Negative values (other than -512)
        SupImCover.AddBins( "sup.im", GenBin (-512, -512, 1) );  -- Special value for 511

        -- Define cross coverage bins
        -- For each of the four inputs, define a number of bins. All the
        -- possibilities will be combined to generate a huge number of bins

        -- Option 1: positive/negative combinations for all 4 inputs
        -- 2 bins ^ 4 inputs = 16 combinations
        --CrossCover.AddCross(
        --    GenBin (-512, 511, 2),
        --    GenBin (-512, 511, 2),
        --    GenBin (-512, 511, 2),
        --    GenBin (-512, 511, 2)
        --);

        -- Option 2: use the same bins we defined for each of the individual
        -- inputs. Bins are concatenated with the & operator.
        -- 5 bins ^ 4 inputs = 625 possibilities
        CrossCover.AddCross (
            GenBin ( 511,  511, 1) & GenBin (   1,  510, 1) & GenBin (   0, 0, 1) & GenBin (-511,   -1, 1) & GenBin (-512, -512, 1),
            GenBin ( 511,  511, 1) & GenBin (   1,  510, 1) & GenBin (   0, 0, 1) & GenBin (-511,   -1, 1) & GenBin (-512, -512, 1),
            GenBin ( 511,  511, 1) & GenBin (   1,  510, 1) & GenBin (   0, 0, 1) & GenBin (-511,   -1, 1) & GenBin (-512, -512, 1),
            GenBin ( 511,  511, 1) & GenBin (   1,  510, 1) & GenBin (   0, 0, 1) & GenBin (-511,   -1, 1) & GenBin (-512, -512, 1)
        );

        -- Record the data in the Coverage Objects every time we have a valid input
        while clk_enable = true loop

        -- We could use "while not CoverageObj.IsCovered" in order to generate
        -- stimuli until we cover everything! In that case we should have a way
        -- to communicate with the stimulus generator (instead of generating a
        -- fixed number of stimuli, just generate random stimuli until desired
        -- coverage is reached)

            wait on clk until clk = '1' and valid = '1';
            InfReCover.ICover(to_integer(inf.re));
            InfImCover.ICover(to_integer(inf.im));
            SupReCover.ICover(to_integer(sup.re));
            SupImCover.ICover(to_integer(sup.im));
            CrossCover.ICover((to_integer(inf.re), to_integer(inf.im), to_integer(sup.re), to_integer(sup.im)));

        end loop;

    end process;

end software;
