library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use work.edc_common.all;

entity tb_interpolator is
end tb_interpolator;

architecture software of tb_interpolator is

    -- I/Os
    signal rst, clk           : std_logic;
    signal valid, estim_valid : std_logic;
    signal inf, sup, estim    : complex10 := (re => to_signed(0,10), im => to_signed(0, 10));

    -- clk control
    signal enable_clk   : boolean := true;
    signal cycle_count  : integer := 0;
    constant CLK_PERIOD : time := 10 ns;
    constant NUM_CYCLES : integer := 120;

    -- protocol monitor
    signal monitor_i      : integer := 12;
    signal expected_re    : real;
    signal expected_im    : real;
    signal expected       : complex10;

begin

    interpolator_inst: entity work.interpolator
    port map (
                rst         => rst,
                clk         => clk,
                inf         => inf,
                sup         => sup,
                valid       => valid,
                estim       => estim,
                estim_valid => estim_valid
             );

    -- Substitute these two processes for the UVVM clock generator
    clk_control: process(clk)
    begin
        if rising_edge(clk) then
            if cycle_count < NUM_CYCLES then
                enable_clk <= true;
                cycle_count <= cycle_count + 1;
            else
                enable_clk <= false;
            end if;
        end if;
    end process;

    clk_process: process
    begin
        if (enable_clk) then
	        clk <= '0';
	        wait for CLK_PERIOD*1/2;
	        clk <= '1';
	        wait for CLK_PERIOD*1/2;
        else
                wait;  -- Stop sim by clk asphyxiation
        end if;
    end process;

    -- Substitute this for the main test control process
    -- Look into generating random inputs
    stim_process: process
    begin
        rst <= '1';
        valid <= '0';
        inf <= (re => (others=>'0'),     im => (others=>'0'));
        sup <= (re => (others=>'0'),     im => (others=>'0'));
        wait for 1*clk_period;
        rst <= '0';
        valid <= '1';
        sup.re <= to_signed(500, sup.re'length);
        sup.im <= to_signed(-500, sup.im'length);
        wait for 1*clk_period;
        valid <= '0';
        wait for 12*clk_period;
        valid <= '1';
        sup.re <= to_signed(250, sup.re'length);
        sup.im <= to_signed(-250, sup.im'length);
        wait for 1*clk_period;
        valid <= '0';

        wait;
    end process;

    -- Monitor that the outputs are correct
    protocol_monitor: process (clk)
    begin
        if rising_edge(clk) then
            if (valid = '1') then
                if (monitor_i = 12) then
                    monitor_i <= 0;
                else
                    report "unexpected assertion of valid input" severity failure;
                end if;
            end if;
            if (estim_valid = '1') then
                -- Here we need UVVM's check_data capabilities, because
                -- printing expected and actual is fairly tedious 
                if expected /= estim then
                    report "error in interpolated data. monitor_i: " &
                    integer'image(monitor_i) & " " &
                    "expected: (" &
                    integer'image(to_integer(expected.re)) & ", " &
                    integer'image(to_integer(expected.im)) & ") " &
                    "actual: (" &
                    integer'image(to_integer(estim.re)) & ", " &
                    integer'image(to_integer(estim.im)) & ")"
                    severity error;
                else
                    report "interpolated data ok" severity note;
                end if;
                monitor_i <= monitor_i + 1;
            end if;
        end if;
    end process;

    -- Calculate the expected interpolated data in a different way than in the
    -- interpolator. We don't want to replicate that same code!
    -- So: compute the equation of the straight line and scale by the scale
    -- factor (*12 because we estimate 12*y, and /16 because we drop the 4 LSB)
    expected_re <= (real(12)/real(16))*((real(to_integer(sup.re) - to_integer(inf.re))/real(12))*real(monitor_i) + real(to_integer(inf.re)));
    expected_im <= (real(12)/real(16))*((real(to_integer(sup.im) - to_integer(inf.im))/real(12))*real(monitor_i) + real(to_integer(inf.im)));
    expected.re <= to_signed(integer((expected_re)), 10);
    expected.im <= to_signed(integer((expected_im)), 10);

end software;
